import os
from pathlib import Path
from scipy import signal
import numpy as np
import pandas as pd
from PyQt5.QtWidgets import QMessageBox
from PyQt5 import QtGui
import pyqtgraph as pg
import datetime
import pynwb
import ndx_ecog


class TimeSeriesPlotter:
    """
    This class holds the 3 time series subplots in the main window.
    It also holds information of the currently open NWB file, as well as user
    annotations and intervals.
    """
    def __init__(self, par):
        self.parent = par
        self.source_path = Path(par.source_path)

        # Makes nwbfile object from nwb file
        if self.source_path.is_file():
            if '_' in self.source_path.name:
                self.subject_id = self.source_path.name.split('_')[0][2:]
                self.block = self.source_path.name.split('_')[1][1:-4]
            else:
                self.subject_id = ''
                self.block = ''
            self.io = pynwb.NWBHDF5IO(str(self.source_path), 'r+', load_namespaces=True)
            self.nwb = self.io.read()      # reads NWB file
        else:
            raise TypeError("Invalid file type")

        title = ' '.join(['EcogVIS -', self.parent.current_session,
                          self.source_path.name, '-', self.block])
        self.parent.setWindowTitle(title)

        # Tries to load Raw data
        lis = list(self.nwb.acquisition.keys())
        there_is_raw = False
        for i in lis:  # Check if there is ElectricalSeries in acquisition group
            if type(self.nwb.acquisition[i]).__name__ == 'ElectricalSeries':
                self.source = self.nwb.acquisition[i]
                self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('raw'))
                there_is_raw = True
        if not there_is_raw:
            print("No 'ElectricalSeries' object in 'acquisition' group.")
        # Tries to load preprocessed data
        try:
            self.source = self.nwb.processing['ecephys'].data_interfaces['LFP'].electrical_series['preprocessed']
            self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('preprocessed'))
            self.parent.push5_0.setEnabled(True)
            self.parent.push6_0.setEnabled(False)
            self.parent.push7_0.setEnabled(True)
        except:
            print("No 'preprocessed' data in 'processing' group.")
        # Tries to load High Gamma data
        try:
            self.source = self.nwb.processing['ecephys'].data_interfaces['high_gamma']
            self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('high gamma'))
            self.parent.push5_0.setEnabled(False)
            self.parent.push6_0.setEnabled(False)
            self.parent.push7_0.setEnabled(False)
        except:
            print("No 'high_gamma' data in 'processing' group.")

        self.plotData = self.source.data
        self.fs_signal = self.source.rate     # sampling frequency [Hz]
        self.tbin_signal = 1 / self.fs_signal  # time bin duration [seconds]
        self.nBins = self.source.data.shape[0]     # total number of bins
        self.min_window_bins = 10                   # minimum number of bins to plot
        # Electrodes table - bipolar or regular table
        self.electrodes_table = self.source.electrodes.table
        # all electricalseries channels ids
        self.all_channels_ids = self.electrodes_table.id[:]
        self.electrical_series_channel_ids = np.array(self.all_channels_ids)[self.source.electrodes.data[:]].tolist()
        self.n_channels_total = len(self.electrical_series_channel_ids)     # total number of channels

        # Get Brain regions present in current file
        self.all_regions = list(self.electrodes_table['location'][self.electrical_series_channel_ids])
        self.all_regions.sort()
        self.regions_mask = [True] * len(self.all_regions)

        self.channels_mask = np.ones(len(list(self.electrodes_table['location'][self.electrical_series_channel_ids])))
        self.channels_mask_ind = np.where(self.channels_mask)[0]

        self.h = []
        self.text = []
        self.AnnotationsList = []
        self.AnnotationsPosAV = np.array([])    # (x, y_va, y_off)
        self.unsaved_changes_annotation = False
        self.unsaved_changes_interval = False

        # Channels to show
        self.firstCh = int(self.parent.qline1.text())
        self.lastCh = min(self.n_channels_total, int(self.parent.qline0.text()))
        self.parent.qline0.setText(str(self.lastCh))
        self.nChToShow = self.lastCh - self.firstCh + 1
        self.selectedChannels = np.arange(self.firstCh - 1, self.lastCh)

        self.current_rect = []

        # List of bad channels
        if 'bad' in self.electrodes_table:
            aux_mask = self.electrodes_table[self.electrical_series_channel_ids]['bad']
            self.bad_channels_ids = list(self.electrodes_table[self.electrical_series_channel_ids][aux_mask].index)
        else:
            self.bad_channels_ids = []

        # Load invalid intervals from NWB file
        self.allIntervals = []
        if self.nwb.invalid_times is not None:
            self.nBI = self.nwb.invalid_times.columns[0][:].shape[0]  # number of BI
            for ii in np.arange(self.nBI):
                # create new interval
                obj = CustomInterval()
                obj.start = self.nwb.invalid_times.columns[0][ii]
                obj.stop = self.nwb.invalid_times.columns[1][ii]
                obj.type = 'invalid'
                obj.color = 'red'
                obj.session = ''
                self.allIntervals.append(obj)

        # Plot interval rectangles at upper and middle pannels
        self.IntRects1 = np.array([], dtype='object')
        self.IntRects2 = np.array([], dtype='object')
        for i, obj in enumerate(self.allIntervals):
            start = obj.start
            stop = obj.stop
            # on timeline
            c = pg.QtGui.QGraphicsRectItem(start, -1, max(stop - start, 0.01), 2)
            self.IntRects1 = np.append(self.IntRects1, c)
            c.setPen(pg.mkPen(color='r'))
            c.setBrush(QtGui.QColor(255, 0, 0, 120))
            self.parent.win2.addItem(c)
            # on signals plot
            c = pg.QtGui.QGraphicsRectItem(start, -1, max(stop - start, 0.01), 2)
            self.IntRects2 = np.append(self.IntRects2, c)
            c.setPen(pg.mkPen(color='r'))
            c.setBrush(QtGui.QColor(255, 0, 0, 120))
            self.parent.win1.addItem(c)

        # Test if current nwb file contains Survey table
        if 'behavior' in self.nwb.processing:
            list_surveys = [v for v in self.nwb.processing['behavior'].data_interfaces.values()
                            if v.neurodata_type == 'SurveyTable']
            if len(list_surveys) > 0:
                self.parent.action_vis_survey.setEnabled(True)
                self.parent.action_add_survey.setEnabled(False)
            else:
                self.parent.action_vis_survey.setEnabled(False)
                self.parent.action_add_survey.setEnabled(True)
        else:
            self.parent.action_vis_survey.setEnabled(False)
            self.parent.action_add_survey.setEnabled(True)

        # Test if current nwb file contains Transcription tables
        if 'behavior' in self.nwb.processing:
            list_transcripts = [v for v in self.nwb.processing['behavior'].data_interfaces.values()
                                if v.name in ['phonemes', 'syllables', 'words', 'sentences']]
            if len(list_transcripts) > 0:
                self.parent.transcriptionadd_tools_menu.setEnabled(False)
                self.parent.action_vis_transcription.setEnabled(True)
            else:
                self.parent.transcriptionadd_tools_menu.setEnabled(True)
                self.parent.action_vis_transcription.setEnabled(False)
        else:
            self.parent.transcriptionadd_tools_menu.setEnabled(True)
            self.parent.action_vis_transcription.setEnabled(False)

        # Add Speaker and Mic Intervals if they exist
        self.SpeakerAndMicIntervalAdd()

        # Load stimuli signals (audio)
        self.load_stimuli()

        # Initiate plots
        self.updateCurXAxisPosition()
        self.refreshScreen()

    def load_stimuli(self):
        """Loads stimuli signals (speaker audio)."""
        self.nStim = len(self.nwb.stimulus)
        self.stimList = list(self.nwb.stimulus.keys())
        self.stimX = []
        self.stimY = {}
        self.parent.combo4.clear()
        for stim in self.stimList:
            self.parent.combo4.addItem(stim)   # add stimulus name to dropdown button
            dt_stim = 1. / self.nwb.stimulus[stim].rate
            nb_stim = self.nwb.stimulus[stim].data.shape[0]
            self.stimX = np.linspace(dt_stim, nb_stim * dt_stim, nb_stim)
            self.stimY[stim] = self.nwb.stimulus[stim].data
        else:
            self.disp_audio = 0

    def refresh_file(self):
        """Re-opens the current file, for when new data is included"""
        if hasattr(self, 'io'):
            self.io.close()   # closes current NWB file

        self.io = pynwb.NWBHDF5IO(str(self.source_path), 'r+', load_namespaces=True)
        self.nwb = self.io.read()      # reads NWB file

        # Searches for signal source on file
        try:   # Tries to load Raw data
            lis = list(self.nwb.acquisition.keys())
            for i in lis:  # Check if there is ElectricalSeries in acquisition group
                if type(self.nwb.acquisition[i]).__name__ == 'ElectricalSeries':
                    self.source = self.nwb.acquisition[i]
                    self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('raw'))
        except:
            None
        try:  # Tries to load preprocessed data
            self.source = self.nwb.processing['ecephys'].data_interfaces['LFP'].electrical_series['preprocessed']
            self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('preprocessed'))
        except:
            None
        try:  # Tries to load High Gamma data
            self.source = self.nwb.processing['ecephys'].data_interfaces['high_gamma']
            self.parent.combo3.setCurrentIndex(self.parent.combo3.findText('high gamma'))
        except:
            None
        self.plotData = self.source.data
        self.fs_signal = self.source.rate      # sampling frequency [Hz]
        self.tbin_signal = 1 / self.fs_signal  # time bin duration [seconds]
        self.nBins = self.source.data.shape[0]     # total number of bins
        self.load_stimuli()  # load stimuli signals (audio)
        self.updateCurXAxisPosition()

    def refreshScreen(self):
        """Re-draws all plots"""
        self.TimeSeries_plotter()

    def TimeSeries_plotter(self):
        """Plots time series signals"""
        startSamp = self.intervalStartSamples
        endSamp = self.intervalEndSamples

        # Bins to plot - subsample for too big arrays
        maxBins = 1000
        ratio_to_max = (endSamp - startSamp) / maxBins
        if ratio_to_max > 1:
            bins_to_plot = np.linspace(startSamp, endSamp, maxBins, dtype='int')
        else:
            bins_to_plot = np.arange(startSamp, endSamp, dtype='int')

        # Use the same scaling factor for all channels, to keep things comparable
        self.verticalScaleFactor = float(self.parent.qline4.text())
        scaleFac = 2 * np.std(self.plotData[startSamp:endSamp, self.selectedChannels - 1], axis=0) / self.verticalScaleFactor

        # Scale variance_units, offset for each channel
        scale_va = np.max(scaleFac)
        self.scaleVec = np.arange(1, self.nChToShow + 1) * scale_va

        scaleV = np.zeros([len(self.scaleVec), 1])
        scaleV[:, 0] = self.scaleVec

        # constrains the plotData to the chosen interval (and transpose matix)
        # plotData dims=[self.nChToShow, plotInterval]
        data = self.plotData[startSamp:endSamp, self.selectedChannels - 1].T
        data = data[:, bins_to_plot - startSamp - 1]
        means = np.reshape(np.mean(data, 1), (-1, 1))  # to align each trace around its reference trace
        plotData = data + scaleV - means  # data + offset

        # Middle signals plot
        # A line indicating reference for every channel
        timebaseGuiUnits = bins_to_plot * self.tbin_signal
        # timebaseGuiUnits = np.arange(startSamp, endSamp)*self.tbin_signal
        plt2 = self.parent.win1  # middle signal plot
        plt2.clear()
        # Channels reference lines
        for l in range(len(self.scaleVec)):
            plt2.plot([timebaseGuiUnits[0], timebaseGuiUnits[-1]],
                      [self.scaleVec[l], self.scaleVec[l]],
                      pen='k')

        # Iterate over chosen channels, plot one at a time
        nrows, ncols = np.shape(plotData)
        for i in range(nrows):
            elec_index = self.source.electrodes[int(self.selectedChannels[i])].index[0]
            if elec_index in self.bad_channels_ids:
                plt2.plot(timebaseGuiUnits, plotData[i], pen=pg.mkPen((220, 0, 0), width=1.2))
            else:
                c = pg.mkPen((0, 120, 0), width=1.2)
                if i % 2 == 0:
                    c = pg.mkPen((0, 0, 200), width=1.2)
                plt2.plot(timebaseGuiUnits, plotData[i], pen=c)
        plt2.setLabel('bottom', 'Time', units='sec')
        plt2.setLabel('left', 'Channel #')
        labels = [str(self.electrical_series_channel_ids[ch]) for ch in self.selectedChannels]
        ticks = list(zip(self.scaleVec, labels))
        plt2.getAxis('left').setTicks([ticks])
        plt2.setXRange(timebaseGuiUnits[0], timebaseGuiUnits[-1], padding=0.003)
        plt2.setYRange(self.scaleVec[0], self.scaleVec[-1], padding=0.06)
        plt2.getAxis('left').setWidth(w=53)

        # Show Intervals
        for i in range(len(self.IntRects2)):
            plt2.removeItem(self.IntRects2[i])
        for i in range(len(self.IntRects2)):
            plt2.addItem(self.IntRects2[i])

        # Show Annotations
        for i in range(len(self.AnnotationsList)):
            plt2.removeItem(self.AnnotationsList[i])
        for i in range(len(self.AnnotationsList)):
            aux = self.AnnotationsList[i].pg_item
            x = self.AnnotationsPosAV[i, 0]
            # Y to plot = (Y_va + Channel offset)*scale_variance
            y_va = self.AnnotationsPosAV[i, 1]
            y = (y_va + self.AnnotationsPosAV[i, 2] - self.firstCh) * scale_va
            aux.setPos(x, y)
            plt2.addItem(aux)

        # Upper horizontal bar
        plt1 = self.parent.win2
        plt1.clear()
        max_dur = self.nBins * self.tbin_signal
        plt1.plot([0, max_dur], [0, 0], pen=pg.mkPen('k', width=2))
        plt1.setXRange(0, max_dur)
        plt1.setYRange(-1, 1)

        if self.current_rect != []:
            plt1.removeItem(self.current_rect)

        # Rectangle Plot
        x = float(self.parent.qline2.text())
        w = float(self.parent.qline3.text())
        self.current_rect = CustomBox(self, x, -1000, w, 2000)
        self.current_rect.setPen(pg.mkPen(color=(0, 0, 0, 50)))
        self.current_rect.setBrush(QtGui.QColor(0, 0, 0, 50))
        self.current_rect.setFlags(QtGui.QGraphicsItem.ItemIsMovable)
        plt1.addItem(self.current_rect)
        plt1.setLabel('left', 'Span')
        plt1.getAxis('left').setWidth(w=53)
        plt1.getAxis('left').setStyle(showValues=False)
        plt1.getAxis('left').setTicks([])

        # Show Intervals
        for i in range(len(self.IntRects1)):
            plt1.removeItem(self.IntRects1[i])
        for i in range(len(self.IntRects1)):
            plt1.addItem(self.IntRects1[i])

        # Bottom plot - Stimuli
        plt3 = self.parent.win3
        plt3.clear()
        stimData = None
        if self.parent.combo4.currentText() != '':
            try:
                xmask = (self.stimX > timebaseGuiUnits[0]) * (self.stimX < timebaseGuiUnits[-1])
                stimName = self.parent.combo4.currentText()
                stimData = self.stimY[stimName]
                plt3.plot(self.stimX[xmask], stimData[xmask], pen='k', width=1)
                plt3.setYRange(np.min(stimData), np.max(stimData))
                plt3.setXLink(plt2)
            except:  # THIS IS MOMENTARY TO PLOT ARTIFICIAL DATA-- REMOVE LATER
                stimName = self.parent.combo4.currentText()
                stimData = self.stimY[stimName]
                plt3.plot(stimData, pen='k', width=1)
                # remove this later -------------------------------------------

        if stimData is not None:
            plt3.setYRange(np.min(stimData), np.max(stimData))
        plt3.setLabel('left', 'Stim')
        plt3.getAxis('left').setWidth(w=53)
        plt3.getAxis('left').setStyle(showValues=False)
        plt3.getAxis('left').setTicks([])

        # set colors
        plt1.getAxis('left').setPen(pg.mkPen(color=(50, 50, 50)))
        plt2.getAxis('left').setPen(pg.mkPen(color=(50, 50, 50)))
        plt2.getAxis('bottom').setPen(pg.mkPen(color=(50, 50, 50)))
        plt3.getAxis('left').setPen(pg.mkPen(color=(50, 50, 50)))

    def drag_window(self, dt):
        """Updates upper visualization window position when dragged by the user."""
        self.parent.qline2.setText(str(self.intervalStartGuiUnits + dt))
        self.updateCurXAxisPosition()

    def updateCurXAxisPosition(self):
        """Updates the time interval selected from the user forms,
           with checks for allowed values"""
        # Read from user forms
        self.intervalStartGuiUnits = float(self.parent.qline2.text())
        self.intervalLengthGuiUnits = float(self.parent.qline3.text())

        max_dur = self.nBins * self.tbin_signal  # Max duration in seconds
        min_len = self.min_window_bins * self.tbin_signal          # Min accepted length

        # Check for max and min allowed values: START
        if self.intervalStartGuiUnits < 0.001:
            self.intervalStartGuiUnits = 0.001
        elif self.intervalStartGuiUnits > max_dur - min_len:
            self.intervalStartGuiUnits = max_dur - min_len

        # Check for max and min allowed values: LENGTH
        if self.intervalLengthGuiUnits + self.intervalStartGuiUnits > max_dur:
            self.intervalLengthGuiUnits = max_dur - self.intervalStartGuiUnits
        elif self.intervalLengthGuiUnits < min_len:
            self.intervalLengthGuiUnits = min_len

        # Updates form values (round to miliseconds for display)
        self.intervalStartGuiUnits = np.round(self.intervalStartGuiUnits, 3)
        self.intervalLengthGuiUnits = np.round(self.intervalLengthGuiUnits, 3)
        self.parent.qline2.setText(str(self.intervalStartGuiUnits))
        self.parent.qline3.setText(str(self.intervalLengthGuiUnits))

        # Updates times in samples
        self.intervalLengthSamples = int(np.floor(self.intervalLengthGuiUnits / self.tbin_signal))
        self.intervalStartSamples = max(np.floor(self.intervalStartGuiUnits / self.tbin_signal), 0)
        self.intervalStartSamples = int(min(self.intervalStartSamples, self.nBins - self.min_window_bins))
        self.intervalEndSamples = self.intervalStartSamples + self.intervalLengthSamples

        self.refreshScreen()

    def time_window_resize(self, wscale):
        """Updates the plotting time interval range"""
        self.intervalLengthGuiUnits = float(self.parent.qline3.text())
        self.intervalLengthGuiUnits = self.intervalLengthGuiUnits * wscale
        self.parent.qline3.setText(str(self.intervalLengthGuiUnits))
        self.updateCurXAxisPosition()

    def time_scroll(self, scroll=0):
        """Updates the plotting time interval for scrolling.
           Buttons: >>, >, <<, <"""
        # new interval start
        self.intervalStartSamples = self.intervalStartSamples + int(scroll * self.intervalLengthSamples)
        # if the last sample of the new interval is beyond is the available data
        # set start such that intserval is in a valid range
        if self.intervalStartSamples + self.intervalLengthSamples > self.nBins:
            self.intervalStartSamples = self.nBins - self.intervalLengthSamples
        elif self.intervalStartSamples < 0:
            self.intervalStartSamples = 0

        # new interval end
        self.intervalEndSamples = self.intervalStartSamples + self.intervalLengthSamples
        # Updates start time in seconds (first avoid <0.001 values)
        self.intervalStartGuiUnits = max(self.intervalStartSamples * self.tbin_signal, 0.001)
        # Round to miliseconds for display
        self.intervalStartGuiUnits = np.round(self.intervalStartGuiUnits, 3)
        self.intervalEndGuiUnits = self.intervalEndSamples * self.tbin_signal
        self.parent.qline2.setText(str(self.intervalStartGuiUnits))
        self.refreshScreen()

    def channel_Scroll_Up(self, opt='unit'):
        """Updates the channels to be plotted. Buttons: ^, ^^ """
        # Test upper limit
        if self.lastCh < self.n_channels_total:
            if opt == 'unit':
                step = 1
            elif opt == 'page':
                step = np.minimum(self.nChToShow, self.n_channels_total - self.lastCh)
            # Add to first and last channels
            self.firstCh += step
            self.lastCh += step
            self.parent.qline0.setText(str(self.lastCh))
            self.parent.qline1.setText(str(self.firstCh))
            self.nChToShow = self.lastCh - self.firstCh + 1
            self.selectedChannels = self.channels_mask_ind[self.firstCh - 1:self.lastCh]
        self.refreshScreen()

    def channel_Scroll_Down(self, opt='unit'):
        """Updates the channels to be plotted. Buttons: v, vv """
        # Test lower limit
        if self.firstCh > 1:
            if opt == 'unit':
                step = 1
            elif opt == 'page':
                step = np.minimum(self.nChToShow, self.firstCh - 1)
            # Subtract from first and last channels
            self.firstCh -= step
            self.lastCh -= step
            self.parent.qline0.setText(str(self.lastCh))
            self.parent.qline1.setText(str(self.firstCh))
            self.nChToShow = self.lastCh - self.firstCh + 1
            self.selectedChannels = self.channels_mask_ind[self.firstCh - 1:self.lastCh]
        self.refreshScreen()

    def nChannels_Displayed(self):
        """Updates channels to be plotted."""
        self.firstCh = int(self.parent.qline1.text())
        self.lastCh = int(self.parent.qline0.text())

        # Make sure indices are in a valid range
        if self.firstCh < 1:
            self.firstCh = 1

        if self.firstCh > self.n_channels_total:
            self.firstCh = self.n_channels_total
            self.lastCh = self.n_channels_total

        if self.lastCh < 1:
            self.lastCh = self.firstCh

        if self.lastCh > self.n_channels_total:
            self.lastCh = self.n_channels_total

        if self.lastCh - self.firstCh < 1:
            self.lastCh = self.firstCh

        # Update variables and values displayed on forms
        self.nChToShow = self.lastCh - self.firstCh + 1
        self.selectedChannels = self.channels_mask_ind[self.firstCh - 1:self.lastCh]
        self.parent.qline0.setText(str(self.lastCh))
        self.parent.qline1.setText(str(self.firstCh))
        self.refreshScreen()

    # Annotation functions ----------------------------------------------------
    def AnnotationAdd(self, x, y, color='yellow', text=''):
        """
        Adds new annotation to plot scene.

        Parameters
        ----------
        x : float
            x position
        y : float
            y position
        color :
            'yellow', 'red', 'green' or 'blue'
        text : str
            Annotation text
        """
        if color == 'yellow':
            bgcolor = pg.mkBrush(250, 250, 150, 200)
        elif color == 'red':
            bgcolor = pg.mkBrush(250, 0, 0, 200)
        elif color == 'green':
            bgcolor = pg.mkBrush(0, 255, 0, 200)
        elif color == 'blue':
            bgcolor = pg.mkBrush(0, 0, 255, 200)
        c = pg.TextItem(anchor=(.5, .5), border=pg.mkPen(100, 100, 100), fill=bgcolor)
        c.setText(text=text, color=(0, 0, 0))
        # Y coordinate transformed to variance_units (for plot control)
        y_va = np.round(y / self.scaleVec[0]).astype('int')
        c.setPos(x, y_va)
        # create annotation object and add it to the list
        obj = CustomAnnotation()
        obj.pg_item = c
        obj.x = x
        obj.y_va = y_va
        obj.y_off = self.firstCh
        obj.color = color
        obj.text = text
        obj.session = self.parent.current_session
        self.AnnotationsList = np.append(self.AnnotationsList, obj)
        # List of positions (to calculate distances)
        if len(self.AnnotationsPosAV) > 0:
            self.AnnotationsPosAV = np.concatenate((self.AnnotationsPosAV,
                                                    np.array([x, y_va, self.firstCh]).reshape(1, 3)))
        else:
            self.AnnotationsPosAV = np.array([x, y_va, self.firstCh]).reshape(1, 3)
        self.refreshScreen()
        self.unsaved_changes_annotation = True

    def AnnotationDel(self, x, y):
        """
        Deletes from plot scene the annotation closest to the passed position.

        Parameters
        ----------
        x : float
            x position
        y : float
            y position
        """
        x_ann = self.AnnotationsPosAV[:, 0]
        y_ann = np.round((self.AnnotationsPosAV[:, 1] + self.AnnotationsPosAV[:, 2] - self.firstCh) * self.scaleVec[0])
        # Calculates euclidian distance from click
        euclid_dist = np.sqrt((x_ann - x)**2 + (y_ann - y)**2)
        indmin = np.argmin(euclid_dist)
        # Checks if user intends to delete annotation
        text = self.AnnotationsList[indmin].text
        buttonReply = QMessageBox.question(None,
                                           'Delete Annotation', "Delete the annotation: \n\n" + text + ' ?',
                                           QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if buttonReply == QMessageBox.Yes:
            self.AnnotationsList = np.delete(self.AnnotationsList, indmin, axis=0)
            self.AnnotationsPosAV = np.delete(self.AnnotationsPosAV, indmin, axis=0)
            self.refreshScreen()
            self.unsaved_changes_annotation = True

    def AnnotationSave(self):
        """Saves annotations in an external CSV file."""
        buttonReply = QMessageBox.question(None, ' ', 'Save annotations on external file?',
                                           QMessageBox.No | QMessageBox.Yes)
        if buttonReply == QMessageBox.Yes:
            c0 = self.AnnotationsPosAV[:, 0]     # x
            c1 = self.AnnotationsPosAV[:, 1]     # y_va
            c2 = self.AnnotationsPosAV[:, 2]     # y_off
            c3 = [self.AnnotationsList[i].color for i in range(len(self.AnnotationsList))]
            c4 = [self.AnnotationsList[i].text for i in range(len(self.AnnotationsList))]
            c5 = [self.AnnotationsList[i].session for i in range(len(self.AnnotationsList))]
            d = {'x': c0, 'y_va': c1, 'y_off': c2, 'color': c3, 'text': c4, 'session': c5}
            df = pd.DataFrame(data=d)
            fullfile = os.path.join('annotations_' +
                                    datetime.datetime.today().strftime('%Y-%m-%d') +
                                    '.csv')
            df.to_csv(fullfile, header=True, index=True)
            self.unsaved_changes_annotation = False

    def AnnotationLoad(self, fname=''):
        """Loads annotations from an external CSV file."""
        df = pd.read_csv(fname)
        all_x = df['x'].values
        all_y_va = df['y_va'].values
        all_y_off = df['y_off'].values
        texts = df['text'].tolist()
        colors = df['color'].tolist()
        session = df['session'].tolist()
        for i, txt in enumerate(texts):    # Add loaded annotations to graph
            if colors[i] == 'yellow':
                bgcolor = pg.mkBrush(250, 250, 150, 200)
            elif colors[i] == 'red':
                bgcolor = pg.mkBrush(250, 0, 0, 200)
            elif colors[i] == 'green':
                bgcolor = pg.mkBrush(0, 255, 0, 200)
            elif colors[i] == 'blue':
                bgcolor = pg.mkBrush(0, 0, 255, 200)
            c = pg.TextItem(anchor=(.5, .5), border=pg.mkPen(100, 100, 100), fill=bgcolor)
            c.setText(text=txt, color=(0, 0, 0))
            # Y coordinate transformed to variance_units (for plot control)
            y_va = all_y_va[i]
            x = all_x[i]
            c.setPos(x, y_va)
            obj = CustomAnnotation()
            obj.pg_item = c
            obj.x = all_x[i]
            obj.y_va = all_y_va[i]
            obj.y_off = all_y_off[i]
            obj.color = colors[i]
            obj.text = txt
            obj.session = session[i]
            self.AnnotationsList.append(obj)
            if len(self.AnnotationsPosAV) > 0:
                self.AnnotationsPosAV = np.concatenate((self.AnnotationsPosAV,
                                                        np.array([x, y_va, all_y_off[i]]).reshape(1, 3)))
            else:
                self.AnnotationsPosAV = np.array([x, y_va, all_y_off[i]]).reshape(1, 3)
        self.refreshScreen()

    # Interval functions ------------------------------------------------------
    def IntervalAdd(self, interval, int_type, color, session):
        """
        Adds new interval to plot scene.

        Parameters
        ----------
        interval : list of floats
            List with [x_initial, x_final] points.
        int_type : str
            Type of the interval (e.g. 'invalid').
        color :
            'yellow', 'red', 'green' or 'blue'.
        session : str
            Session name.
        """
        if color == 'yellow':
            bc = [250, 250, 150, 180]
        elif color == 'red':
            bc = [250, 0, 0, 100]
        elif color == 'green':
            bc = [0, 255, 0, 130]
        elif color == 'blue':
            bc = [0, 0, 255, 100]
        x = interval[0]
        w = np.diff(np.array(interval))[0]
        # add rectangle to upper plot
        c = pg.QtGui.QGraphicsRectItem(x, -1, w, 2)
        c.setPen(pg.mkPen(color=QtGui.QColor(bc[0], bc[1], bc[2], 255)))
        c.setBrush(QtGui.QColor(bc[0], bc[1], bc[2], bc[3]))
        self.IntRects1 = np.append(self.IntRects1, [c])
        # add rectangle to middle signal plot
        c = pg.QtGui.QGraphicsRectItem(x, -1, w, 2)
        c.setPen(pg.mkPen(color=QtGui.QColor(bc[0], bc[1], bc[2], 255)))
        c.setBrush(QtGui.QColor(bc[0], bc[1], bc[2], bc[3]))
        self.IntRects2 = np.append(self.IntRects2, [c])
        # new Interval object
        obj = CustomInterval()
        obj.start = interval[0]
        obj.stop = interval[1]
        obj.type = int_type
        obj.color = color
        obj.session = session
        self.allIntervals.append(obj)
        self.nBI = len(self.allIntervals)
        self.unsaved_changes_interval = True

    def IntervalDel(self, x):
        """
        Deletes from plot scene the interval passing by the x position.

        Parameters
        ----------
        x : float
            x position.
        """
        for i, obj in enumerate(self.allIntervals):
            if (x >= obj.start) & (x <= obj.stop):   # interval of the click
                self.parent.win2.removeItem(self.IntRects1[i])
                self.IntRects1 = np.delete(self.IntRects1, i, axis=0)
                self.parent.win1.removeItem(self.IntRects2[i])
                self.IntRects2 = np.delete(self.IntRects2, i, axis=0)
                del self.allIntervals[i]
                self.nBI = len(self.allIntervals)
                self.refreshScreen()
                self.unsaved_changes_interval = True

    def IntervalSave(self):
        """Saves intervals in an external CSV file."""
        buttonReply = QMessageBox.question(None, ' ', 'Save intervals on external file?',
                                           QMessageBox.No | QMessageBox.Yes)
        if buttonReply == QMessageBox.Yes:
            c0, c1, c2, c3, c4 = [], [], [], [], []
            for i, obj in enumerate(self.allIntervals):
                c0.append(obj.start)     # start
                c1.append(obj.stop)      # stop
                c2.append(obj.type)      # type
                c3.append(obj.color)     # color
                c4.append(obj.session)   # session
            d = {'start': c0, 'stop': c1, 'type': c2, 'color': c3, 'session': c4}
            df = pd.DataFrame(data=d)
            fullfile = os.path.join('intervals_' +
                                    datetime.datetime.today().strftime('%Y-%m-%d') +
                                    '.csv')
            df.to_csv(fullfile, header=True, index=True)
            self.unsaved_changes_interval = False

    def IntervalLoad(self, fname):
        """Loads intervals from an external CSV file."""
        df = pd.read_csv(fname)
        for index, row in df.iterrows():    # Add loaded intervals to graph
            interval = [row.start, row.stop]
            int_type = row.type
            color = row.color
            session = row.session
            self.IntervalAdd(interval, int_type, color, session)
            # Update dictionary of interval types
            # TO-DO
        self.refreshScreen()

    def SpeakerAndMicIntervalAdd(self):
        if self.nwb.intervals is not None:
            keys = ['TimeIntervals_speaker', 'TimeIntervals_mic']
            colors = ['blue', 'green']
            for name, color in zip(keys, colors):
                if name in self.nwb.intervals:
                    ti = self.nwb.intervals[name]
                    nI = len(ti)
                    for ii in np.arange(nI):
                        interval = [ti['start_time'][ii], ti['stop_time'][ii]]
                        int_type = name
                        color = color
                        session = ''
                        self.IntervalAdd(interval, int_type, color, session)

    # Channels functions -------------------------------------------------------
    def BadChannelAdd(self, ch_list):
        """
        Marks specific channels as 'bad'.

        Parameters
        ----------
        ch_list : list of integers
            List of indices of channels to be marked as 'bad'.
        """
        # Update list of bad electrodes ids
        for ch in ch_list:
            if ch not in self.bad_channels_ids:
                self.bad_channels_ids.append(ch)
        self.update_bad_channels()

    def BadChannelDel(self, ch_list):
        """
        Un-marks specific channels as 'bad'.

        Parameters
        ----------
        ch_list : list of integers
            List of indices of channels to be un-marked as 'bad'.
        """
        # Update list of bad electrodes ids
        for ch in ch_list:
            if ch in self.bad_channels_ids:
                self.bad_channels_ids.remove(ch)
        self.update_bad_channels()

    def update_bad_channels(self):
        """Updates list of bad channels after add or del"""
        # List of electrodes IDs
        elecs_ids = list(self.electrodes_table.id[:])
        is_bad_list = [False] * len(elecs_ids)
        for i, id in enumerate(elecs_ids):
            if id in self.bad_channels_ids:
                is_bad_list[i] = True

        if 'bad' not in self.electrodes_table:
            self.electrodes_table.add_column(
                name='bad',
                description='electrode identified as too noisy',
                data=is_bad_list,
            )
        else:
            self.electrodes_table['bad'].data[:] = is_bad_list

        # Refresh screen
        self.refreshScreen()

    def DrawMarkTime(self, position):
        """Marks temporary reference line when adding a new interval."""
        plt = self.parent.win1     # middle signal plot
        self.current_mark = pg.QtGui.QGraphicsRectItem(position, 0, 0, 1)
        self.current_mark.setPen(pg.mkPen(color='k'))
        self.current_mark.setBrush(QtGui.QColor(150, 150, 150, 100))
        plt.addItem(self.current_mark)

    def RemoveMarkTime(self):
        """Removes temporary reference line when adding a new interval."""
        plt = self.parent.win1     # middle signal plot
        plt.removeItem(self.current_mark)

    def close_nwbfile(self):
        """Close current nwbfile"""
        if hasattr(self, 'io'):
            self.io.close()


class CustomInterval:
    """
    Stores information about individual Intervals.

    Parameters
    ----------
    start : float
        Start time position.
    stop : float
        Stop time position.
    type : str
        Interval type (e.g. 'invalid').
    color : str
        Interval color: 'yellow', 'red', 'green' or 'blue'.
    session: str
        Session name.
    """
    def __init__(self):
        self.start = -999
        self.stop = -999
        self.type = ''
        self.color = ''
        self.session = ''


class CustomAnnotation:
    """
    Stores information about individual Annotations.

    Parameters
    ----------
    pg_item :
        pyqtgraph TextItem object.
    x : float
        X position (time).
    y_va : int
        Y coordinate transformed to variance_units (for plot control).
    y_off : int
        Y offset, usually the bottom channel on plot when annotation is saved.
    color : str
        Interval color: 'yellow', 'red', 'green' or 'blue'.
    text : str
        Annotation text.
    session: str
        Session name.
    """
    def __init__(self):
        self.pg_item = None
        self.x = 0
        self.y_va = 0
        self.y_off = 0
        self.color = ''
        self.text = ''
        self.session = ''


class CustomBox(pg.QtGui.QGraphicsRectItem):
    """Upper visualization window rectangle that can be dragged by the user."""
    def __init__(self, parent, x, y, w, h):
        pg.QtGui.QGraphicsRectItem.__init__(self, x, y, w, h)
        self.parent = parent

    def mouseReleaseEvent(self, event):
        dt = self.x()
        self.parent.drag_window(dt)



from __future__ import print_function, division

import time
import os
import numpy as np
import warnings

from pynwb import NWBHDF5IO, ProcessingModule
from pynwb.ecephys import LFP, ElectricalSeries
from pynwb.core import DynamicTable, VectorData
from pynwb.misc import DecompositionSeries

from ecogvis.signal_processing.hilbert_transform import hilbert_transform
from process_nwb.wavelet_transform import gaussian
from process_nwb.resample import resample
from process_nwb.linenoise_notch import apply_linenoise_notch
from ecogvis.signal_processing.common_referencing import subtract_CAR
from ecogvis.functions.nwb_copy_file import nwb_copy_file


def processing_data(path, subject, blocks, mode=None, config=None, new_file=''):
    for block in blocks:
        block_path = os.path.join(path, '{}_B{}.nwb'.format(subject, block))
        if new_file != '':
            make_new_nwb(old_file=block_path, new_file=new_file)

        if mode == 'preprocess':
            preprocess_raw_data(block_path, config=config)
        elif mode == 'decomposition':
            spectral_decomposition(block_path, bands_vals=config)
        elif mode == 'high_gamma':
            high_gamma_estimation(block_path, bands_vals=config, new_file=new_file)


def make_new_nwb(old_file, new_file, cp_objs=None):
    """
    Copy all fields, as defined by cp_objs, from current NWB file to new NWB
    file.

    Parameters
    ----------
    old_file : str, path
        String such as '/path/to/old_file.nwb'.
    new_file : str, path
        String such as '/path/to/new_file.nwb'.
    cp_objs : dictionary
        Dictionary of objectives to copy from the current NWB file to the
        new NWB file. If not given, the defaults will be set (defined
        below). For 'acquisition', 'stimulus', and 'ecephys' fields, if the
        value is defined as 'default', the default fields from those fields
        will be copied. Otherwise, if specified as a list of strings,
        those specific fields will be copied. If something is not to be
        copied, it should be left out of the cp_objs dictionary.
    """

    # Dictionary for copy_nwb
    if cp_objs is None:
        cp_objs = {
            'institution': True,
            'lab': True,
            'session': True,
            'devices': True,
            'electrode_groups': True,
            'electrodes': True,
            'intervals': True,
            'stimulus': True,
            'subject': True,
            'acquisition': 'default'
        }

    # Open original signal file
    with NWBHDF5IO(old_file, 'r', load_namespaces=True) as io:
        nwb_old = io.read()

        if ('acquisition' in cp_objs) and \
                (cp_objs['acquisition'] == 'default'):
            # If 'acquisition' variables are in cp_objs but not defined,
            # default: Exclude raw signals, keep other acquisition
            acq_vars = list(nwb_old.acquisition.keys())
            if 'ElectricalSeries' in acq_vars:
                acq_vars.remove('ElectricalSeries')
            cp_objs['acquisition'] = acq_vars

        if ('ecephys' in cp_objs) and \
                (cp_objs['ecephys'] == 'default'):
            # If 'ecephys' variables are in cp_objs but not defined, default:
            # Exclude existing high_gamma, keep other processing modules
            ece_vars = list(nwb_old.processing['ecephys'].data_interfaces.keys())
            if 'high_gamma' in ece_vars:
                ece_vars.remove('high_gamma')
            cp_objs['ecephys'] = ece_vars

    nwb_copy_file(old_file, new_file, cp_objs=cp_objs)


def preprocess_raw_data(block_path, config):
    """
    Takes raw data and runs:
    1) CAR
    2) notch filters
    3) Downsampling

    Parameters
    ----------
    block_path : str
        subject file path
    config : dictionary
        'referencing' - tuple specifying electrode referencing (type, options)
            ('CAR', N_channels_per_group)
            ('CMR', N_channels_per_group)
            ('bipolar', INCLUDE_OBLIQUE_NBHD)
        'Notch' - Main frequency (Hz) for notch filters (default=60)
        'Downsample' - Downsampling frequency (Hz, default= 400)

    Returns
    -------
    Saves preprocessed signals (LFP) in the current NWB file.
    Only if containers for these data do not exist in the current file.
    """
    subj_path, block_name = os.path.split(block_path)
    block_name = os.path.splitext(block_path)[0]
    start = time.time()

    with NWBHDF5IO(block_path, 'r+', load_namespaces=True) as io:
        nwb = io.read()

        # Storage of processed signals on NWB file ----------------------------
        if 'ecephys' in nwb.processing:
            ecephys_module = nwb.processing['ecephys']
        else:   # creates ecephys ProcessingModule
            ecephys_module = ProcessingModule(
                name='ecephys',
                description='Extracellular electrophysiology data.'
            )
            # Add module to NWB file
            nwb.add_processing_module(ecephys_module)
            print('Created ecephys')

        # LFP: Downsampled and power line signal removed ----------------------
        if 'LFP' in nwb.processing['ecephys'].data_interfaces:
            warnings.warn('LFP data already exists in the nwb file. Skipping preprocessing.')
        else:  # creates LFP data interface container
            lfp = LFP()

            # Data source
            source_list = [acq for acq in nwb.acquisition.values()
                           if type(acq) == ElectricalSeries]
            assert len(source_list) == 1, (
                'Not precisely one ElectricalSeries in acquisition!')
            source = source_list[0]
            nChannels = source.data.shape[1]

            # Downsampling
            if config['Downsample'] is not None:
                print("Downsampling signals to " + str(config['Downsample']) + " Hz.")
                print("Please wait...")
                start = time.time()
                # Note: zero padding the signal to make the length
                # a power of 2 won't help, since resample will further pad it
                # (breaking the power of 2)
                nBins = source.data.shape[0]
                rate = config['Downsample']

                # malloc
                T = int(np.ceil(nBins * rate / source.rate))
                X = np.zeros((source.data.shape[1], T))

                # One channel at a time, to improve memory usage for long signals
                for ch in np.arange(nChannels):
                    # 1e6 scaling helps with numerical accuracy
                    Xch = source.data[:, ch] * 1e6
                    X[ch, :] = resample(Xch, rate, source.rate)
                print('Downsampling finished in {} seconds'.format(
                    time.time() - start))
            else:  # No downsample
                rate = source.rate
                X = source.data[()].T * 1e6

            # re-reference the (scaled by 1e6!) data
            electrodes = source.electrodes
            if config['referencing'] is not None:
                if config['referencing'][0] == 'CAR':
                    print("Computing and subtracting Common Average Reference in "
                          + str(config['referencing'][1])+" channel blocks.")
                    start = time.time()
                    X = subtract_CAR(X, b_size=config['referencing'][1])
                    print('CAR subtract time for {}: {} seconds'.format(
                        block_name, time.time() - start))
                elif config['referencing'][0] == 'bipolar':
                    X, bipolarTable, electrodes = get_bipolar_referenced_electrodes(
                        X, electrodes, rate, grid_step=1)

                    # add data interface for the metadata for saving
                    ecephys_module.add_data_interface(bipolarTable)
                    print('bipolarElectrodes stored for saving in ' + block_path)
                else:
                    print('UNRECOGNIZED REFERENCING SCHEME; ', end='')
                    print('SKIPPING REFERENCING!')

            # Apply Notch filters
            if config['Notch'] is not None:
                print("Applying notch filtering of " + str(config['Notch']) + " Hz")
                # Note: zero padding the signal to make the length a power
                # of 2 won't help, since notch filtering will further pad it
                start = time.time()
                for ch in np.arange(nChannels):
                    # NOTE: apply_linenoise_notch takes a signal that is
                    # (n_timePoints, n_channels). The documentation may be wrong
                    Xch = X[ch, :].reshape(-1, 1)
                    Xch = apply_linenoise_notch(Xch, rate)
                    X[ch, :] = Xch[:, 0]
                print('Notch filter time for {}: {} seconds'.format(
                    block_name, time.time() - start))

            X = X.astype('float32')     # signal (nChannels,nSamples)
            X /= 1e6                    # Scales signals back to volts

            # Add preprocessed downsampled signals as an electrical_series
            referencing = 'None' if config['referencing'] is None else config[
                'referencing'][0]
            notch = 'None' if config['Notch'] is None else str(config['Notch'])
            downs = 'No' if config['Downsample'] is None else 'Yes'
            config_comment = (
                'referencing:' + referencing
                + ', Notch:' + notch
                + ', Downsampled:' + downs
            )

            # create an electrical series for the LFP and store it in lfp
            lfp.create_electrical_series(
                name='preprocessed',
                data=X.T,
                electrodes=electrodes,
                rate=rate,
                description='',
                comments=config_comment
            )
            ecephys_module.add_data_interface(lfp)

            # Write LFP to NWB file
            io.write(nwb)
            print('LFP saved in ' + block_path)


def get_bipolar_referenced_electrodes(
    X, electrodes, rate, grid_size=None, grid_step=1
):
    '''
    Bipolar referencing of electrodes according to the scheme of Dr. John Burke

    Each electrode (with obvious exceptions at the edges) yields two bipolar-
    referenced channels, one for the "right" and one for the "below" neighbors.
    These are returned as an ElectricalSeries.

    Input arguments:
    --------
    X:
        numpy array containing (raw) electrode traces (Nelectrodes x T)
    electrodes:
        DynamicTableRegion containing the metadata for the electrodes whose
        traces are in X
    rate:
        sampling rate of X; for storage in ElectricalSeries
    grid_size:
        numpy array with the two dimensions of the grid (2, )

    Returns:
    --------
    bipolarElectrodes:
        ElectricalSeries containing the bipolar-referenced (pseudo) electrodes
        and associated metadata.  (NB that a, e.g., 16x16 grid yields 960 of
        these pseudo-electrodes.)
    '''

    # set mutable default argument(s)
    if grid_size is None:
        grid_size = np.array([16, 16])

    # malloc
    elec_layout = np.arange(np.prod(grid_size) - 1, -1, -1).reshape(grid_size).T
    elec_layout = elec_layout[::grid_step, ::grid_step]
    grid_size = elec_layout.T.shape  # in case grid_step > 1
    Nchannels = 2 * np.prod(grid_size) - np.sum(grid_size)
    XX = np.zeros((Nchannels, X.shape[1]))

    # create a new dynamic table to hold the metadata
    column_names = ['x', 'y', 'z', 'imp', 'location', 'label', 'bad']
    columns = [
        VectorData(
            name=name,
            description=electrodes.table[name].description)
        for name in column_names
    ]
    bipolarTable = DynamicTable(
        name='bipolar-referenced metadata',
        description=('pseudo-channels derived via John Burke style'
                     ' bipolar referencing'),
        colnames=column_names,
        columns=columns,
    )

    # compute bipolar-ref'd channel and add a new row of metadata to table
    def add_new_channel(iChannel, iElectrode, jElectrode):
        jElectrode = int(jElectrode)

        # "bipolar referencing": the difference of neighboring electrodes
        XX[iChannel, :] = X[iElectrode, :] - X[jElectrode, :]

        # add a row to the table for this new pseudo-electrode
        bipolarTable.add_row(
            {
                'location': '_'.join({
                    electrodes.table['location'][iElectrode],
                    electrodes.table['location'][jElectrode]
                }),
                'label': '-'.join([
                    electrodes.table['label'][iElectrode],
                    electrodes.table['label'][jElectrode]
                ]),
                'bad': (
                    electrodes.table['bad'][iElectrode] or
                    electrodes.table['bad'][jElectrode]
                ),
                **{name: electrodes.table[name][iElectrode]
                   for name in ['x', 'y', 'z', 'imp']},
            }
        )
        return iChannel + 1

    iChannel = 0

    # loop across columns and rows (remembering that grid is transposed)
    for i in range(grid_size[1]):
        for j in range(grid_size[0]):
            if j < grid_size[0] - 1:
                iChannel = add_new_channel(
                    iChannel, elec_layout[i, j], elec_layout[i, j + 1])
            if i < grid_size[1] - 1:
                iChannel = add_new_channel(
                    iChannel, elec_layout[i, j], elec_layout[i + 1, j])

    # create one big region for the entire table
    bipolarTableRegion = bipolarTable.create_region(
        'electrodes', [i for i in range(Nchannels)], 'all bipolar electrodes')

    return XX, bipolarTable, bipolarTableRegion


def spectral_decomposition(block_path, bands_vals):
    """
    Takes preprocessed LFP data and does the standard Hilbert transform on
    different bands. Takes about 20 minutes to run on 1 10-min block.

    Parameters
    ----------
    block_path : str
        subject file path
    bands_vals : [2,nBands] numpy array with Gaussian filter parameters, where:
        bands_vals[0,:] = filter centers [Hz]
        bands_vals[1,:] = filter sigmas [Hz]

    Returns
    -------
    Saves spectral power (DecompositionSeries) in the current NWB file.
    Only if container for this data do not exist in the file.
    """

    # Get filter parameters
    band_param_0 = bands_vals[0, :]
    band_param_1 = bands_vals[1, :]

    with NWBHDF5IO(block_path, 'r+', load_namespaces=True) as io:
        nwb = io.read()
        lfp = nwb.processing['ecephys'].data_interfaces['LFP'].electrical_series['preprocessed']
        rate = lfp.rate

        nBands = len(band_param_0)
        nSamples = lfp.data.shape[0]
        nChannels = lfp.data.shape[1]
        Xp = np.zeros((nBands, nChannels, nSamples))  # power (nBands,nChannels,nSamples)

        # Apply Hilbert transform ---------------------------------------------
        print('Running Spectral Decomposition...')
        start = time.time()
        for ch in np.arange(nChannels):
            Xch = lfp.data[:, ch] * 1e6       # 1e6 scaling helps with numerical accuracy
            Xch = Xch.reshape(1, -1)
            Xch = Xch.astype('float32')     # signal (nChannels,nSamples)
            X_fft_h = None
            for ii, (bp0, bp1) in enumerate(zip(band_param_0, band_param_1)):
                kernel = gaussian(Xch.shape[-1], rate, bp0, bp1)
                X_analytic, X_fft_h = hilbert_transform(Xch, rate, kernel, phase=None, X_fft_h=X_fft_h)
                Xp[ii, ch, :] = abs(X_analytic).astype('float32')
        print('Spectral Decomposition finished in {} seconds'.format(time.time() - start))

        # data: (ndarray) dims: num_times * num_channels * num_bands
        Xp = np.swapaxes(Xp, 0, 2)

        # Spectral band power
        # bands: (DynamicTable) frequency bands that signal was decomposed into
        band_param_0V = VectorData(
            name='filter_param_0',
            description='frequencies for bandpass filters',
            data=band_param_0
        )
        band_param_1V = VectorData(
            name='filter_param_1',
            description='frequencies for bandpass filters',
            data=band_param_1
        )
        bandsTable = DynamicTable(
            name='bands',
            description='Series of filters used for Hilbert transform.',
            columns=[band_param_0V, band_param_1V],
            colnames=['filter_param_0', 'filter_param_1']
        )
        decs = DecompositionSeries(
            name='DecompositionSeries',
            data=Xp,
            description='Analytic amplitude estimated with Hilbert transform.',
            metric='amplitude',
            unit='V',
            bands=bandsTable,
            rate=rate,
            source_timeseries=lfp
        )

        # Storage of spectral decomposition on NWB file ------------------------
        ecephys_module = nwb.processing['ecephys']
        ecephys_module.add_data_interface(decs)
        io.write(nwb)
        print('Spectral decomposition saved in ' + block_path)


def high_gamma_estimation(block_path, bands_vals, new_file=''):
    """
    Takes preprocessed LFP data and calculates High-Gamma power from the
    averaged power of standard Hilbert transform on 70~150 Hz bands.

    Parameters
    ----------
    block_path : str
        subject file path
    bands_vals : [2,nBands] numpy array with Gaussian filter parameters, where:
        bands_vals[0,:] = filter centers [Hz]
        bands_vals[1,:] = filter sigmas [Hz]
    new_file : str
        if this argument is of form 'path/to/new_file.nwb', High Gamma power
        will be saved in a new file. If it is an empty string, '', High Gamma
        power will be saved in the current NWB file.

    Returns
    -------
    Saves High Gamma power (TimeSeries) in the current or new NWB file.
    Only if container for this data do not exist in the file.
    """

    # Get filter parameters
    band_param_0 = bands_vals[0, :]
    band_param_1 = bands_vals[1, :]

    with NWBHDF5IO(block_path, 'r+', load_namespaces=True) as io:
        nwb = io.read()
        lfp = nwb.processing['ecephys'].data_interfaces['LFP'].electrical_series['preprocessed']
        rate = lfp.rate

        nBands = len(band_param_0)
        nSamples = lfp.data.shape[0]
        nChannels = lfp.data.shape[1]
        Xp = np.zeros((nBands, nChannels, nSamples))  # power (nBands,nChannels,nSamples)

        # Apply Hilbert transform ---------------------------------------------
        print('Running High Gamma estimation...')
        start = time.time()
        for ch in np.arange(nChannels):
            Xch = lfp.data[:, ch] * 1e6       # 1e6 scaling helps with numerical accuracy
            Xch = Xch.reshape(1, -1)
            Xch = Xch.astype('float32')     # signal (nChannels,nSamples)
            X_fft_h = None
            for ii, (bp0, bp1) in enumerate(zip(band_param_0, band_param_1)):
                kernel = gaussian(Xch.shape[-1], rate, bp0, bp1)
                X_analytic, X_fft_h = hilbert_transform(
                    Xch, rate, kernel, phase=None, X_fft_h=X_fft_h)
                Xp[ii, ch, :] = abs(X_analytic).astype('float32')
        print('High Gamma estimation finished in {} seconds'.format(time.time() - start))

        # data: (ndarray) dims: num_times * num_channels * num_bands
        Xp = np.swapaxes(Xp, 0, 2)
        HG = np.mean(Xp, 2)   # average of high gamma bands

        # Storage of High Gamma on NWB file -----------------------------
        if new_file == '' or new_file is None:  # on current file
            # make electrodes table
            nElecs = HG.shape[1]
            ecephys_module = nwb.processing['ecephys']

            # first check for a table among the file's data_interfaces
            ####
            if lfp.electrodes.table.name in ecephys_module.data_interfaces:
                LFP_dynamic_table = ecephys_module.data_interfaces[
                    lfp.electrodes.table.name]
            else:
                # othewise use the electrodes as the table
                LFP_dynamic_table = nwb.electrodes
            ####
            elecs_region = LFP_dynamic_table.create_region(
                name='electrodes',
                region=[i for i in range(nChannels)],
                description='all electrodes'
            )
            hg = ElectricalSeries(
                name='high_gamma',
                data=HG,
                electrodes=elecs_region,
                rate=rate,
                description=''
            )

            ecephys_module.add_data_interface(hg)
            io.write(nwb)
            print('High Gamma power saved in ' + block_path)
        else:  # on new file
            with NWBHDF5IO(new_file, 'r+', load_namespaces=True) as io_new:
                nwb_new = io_new.read()
                # make electrodes table
                nElecs = HG.shape[1]
                elecs_region = nwb_new.electrodes.create_region(
                    name='electrodes',
                    region=np.arange(nElecs).tolist(),
                    description='all electrodes'
                )
                hg = ElectricalSeries(
                    name='high_gamma',
                    data=HG,
                    electrodes=elecs_region,
                    rate=rate,
                    description=''
                )

                try:      # if ecephys module already exists
                    ecephys_module = nwb_new.processing['ecephys']
                except:   # creates ecephys ProcessingModule
                    ecephys_module = ProcessingModule(name='ecephys',
                                                      description='Extracellular electrophysiology data.')
                    nwb_new.add_processing_module(ecephys_module)

                ecephys_module.add_data_interface(hg)
                io_new.write(nwb_new)
                print('High Gamma power saved in ' + new_file)



"""
Frequency band information for different types of data processing.
"""
import numpy as np

# Chang lab frequencies
fq_min = 4.0749286538265
fq_max = 200.
scale = 7.
cfs = 2 ** (np.arange(np.log2(fq_min) * scale, np.log2(fq_max) * scale) / scale)
cfs = np.array(cfs)
sds = 10 ** ( np.log10(.39) + .5 * (np.log10(cfs)))
sds = np.array(sds) * np.sqrt(2.)
chang_lab = {'fq_min': fq_min,
             'fq_max': fq_max,
             'scale': scale,
             'cfs': cfs,
             'sds': sds}

# Standard neuro bands
bands = ['theta', 'alpha', 'beta', 'gamma', 'high gamma']
min_freqs = [4., 8., 15., 30., 70.]
max_freqs = [7., 14., 29., 59., 150.]
HG_freq = 200.
neuro = {'bands': bands,
         'min_freqs': min_freqs,
         'max_freqs': max_freqs,
         'HG_freq': HG_freq}



# -*- coding: utf-8 -*-
"""
This code has been modified from David Moses' original version by Ben
Dichter to be stand-alone

-----------------------------------------------------------------------------
Name:        HTK
Purpose:     To interact with .htk files

Author:      David Moses

Created:     Dec 10, 2013
Copyright:   (c) David Moses 2013
"""

import os

import numpy as np

# In the MATLAB function named writeHTK, the sampling rate is multiplied by a
# scaling factor before being written to the file. The reason for doing this is
# unknown. That scaling factor is stored here as a static variable for
# conversion from what the file specifies as the sampling rate into the actual
# sampling rate (in Hz), assuming that the scaling factor was applied during
# the HTK file construction.
SAMPLING_RATE_FACTOR = 1e4


def readHTK(file_path, data_type='HTK', big_endian=True,
            scale_s_rate=False):
    """
    Reads data from a HTK file
    Parameters:
    - file_path     (string)  - The HTK file name
    - data_type     (string)  - Specifies the data type (view getFilePath for
                                more details)
    - big_endian    (Boolean) - Specifies whether or not the data is big endian
                                (if False, then the data is little endian)
    - scale_s_rate  (Boolean) - Specifies whether or note to divide the
                                sampling rate specified by the HTK file by the
                                sampling rate scaling factor
    Returns:
    A dictionary containing the data from the file
    """

    # Creates the endian variable that will be used when specifying data types
    if big_endian:
        endian = '>'
    else:
        endian = '<'

    # Opens the HTK file
    with open(file_path, 'rb') as f:

        # Reads values from the header
        num_samples = np.fromfile(f, dtype=endian + 'i4', count=1)[0]
        sampling_rate = np.fromfile(f, dtype=endian + 'i4', count=1)[0]
        sample_size = np.fromfile(f, dtype=endian + 'i2', count=1)[0]
        parameter_kind = np.fromfile(f, dtype=endian + 'i2', count=1)[0]

        # Reads the data
        data = np.fromfile(f, dtype=endian + 'f4').reshape((num_samples, -1)).T

    # Scales the sampling rate (if desired)
    if scale_s_rate:
        sampling_rate = float(sampling_rate) / SAMPLING_RATE_FACTOR

    # Returns a dictionary containing the data from the file
    return {
        'num_samples': num_samples,
        'sampling_rate': float(sampling_rate),
        'sample_size': sample_size,
        'parameter_kind': parameter_kind,
        'data': data
    }


def toHTK(file_data, file_path, data_type='HTK', big_endian=True,
          scale_s_rate=False):
    """
    Writes data to a HTK file.
    Parameters:
    - file_data     (dict)    - A dictionary containing the following items:
                                - 'num_samples'    : The number of samples
                                - 'sampling_rate'  : The sampling rate (in Hz)
                                - 'sample_size'    : Unknown
                                - 'parameter_kind' : Unknown
                                - 'data'           : The data points
                                The sample size and parameter kind items are
                                optional; the others are required.
    - file_name     (string)  - The HTK file name
    - data_type     (string)  - Specifies the data type (view getFilePath for
                                more details)
    - big_endian    (Boolean) - Specifies whether or not the data is big endian
                                (if False, then the data is little endian)
    - scale_s_rate  (Boolean) - Specifies whether or note to multiply the
                                sampling rate specified by the HTK file by the
                                sampling rate scaling factor
    """

    # Creates the endian variable that will be used when specifying data types
    if big_endian:
        endian = '>'
    else:
        endian = '<'

    # Obtains relevant values from the file data dictionary. If the sample
    # size or the parameter kind keys are not in the dictionary, then
    # arbitrary values are used.
    num_samples = file_data['num_samples']
    sampling_rate = file_data['sampling_rate']
    data = file_data['data']

    try:
        sample_size = file_data['sample_size']
    except KeyError:
        sample_size = 0

    try:
        parameter_kind = file_data['parameter_kind']
    except KeyError:
        parameter_kind = 8971

    # Scales the sampling rate (if desired)
    if scale_s_rate:
        sampling_rate = float(sampling_rate) * SAMPLING_RATE_FACTOR

    # Opens the HTK file
    with open(file_path, 'wb') as f:

        # Writes the header values
        np.array(num_samples, dtype=endian + 'i4').tofile(f)
        np.array(sampling_rate, dtype=endian + 'i4').tofile(f)
        np.array(sample_size, dtype=endian + 'i2').tofile(f)
        np.array(parameter_kind, dtype=endian + 'i2').tofile(f)

        # Writes the data
        np.array(data, dtype=endian + 'f4').T.tofile(f)


def readHTKs(dir_path, electrodes=None, take_mean=True):
    """
    Reads all of the HTK files in a directory, optionally takes the mean, and outputs a
    matrix that contains all of the channels
    together
    Parameters:
    - dir_path       (string)  - Directory that contains HTK files
    - electrodes     (array)   - List of which electrodes to include.
                                 Default is all in dir
    - take_mean      (boolean) - Whether to take the mean on axis 0 for each HTK file
    Returns:
    A dictionary containing the data from the directory
    """
    num_files = len(os.listdir(dir_path))
    if not electrodes:
        electrodes = range(256)

    htkout = readHTK(dir_path + '/Wav11.htk')
    if take_mean:
        num_samples = htkout['data'].shape[1]
        alldata = np.zeros((len(electrodes), num_samples))  # malloc
    else:
        samples_dims = htkout['data'].shape
        alldata = np.zeros([len(electrodes)]+list(samples_dims))


    for i,ifile in enumerate(electrodes):
        htkout = readHTK(dir_path + '/Wav%i%i.htk' % (
        np.ceil((ifile + 1) / 64.), np.mod(ifile, 64) + 1))
        if take_mean:
            alldata[i, :] = np.mean(htkout['data'], 0)
        else:
            alldata[i,] = htkout['data']

    return {
        'num_samples': htkout['num_samples'],
        'sampling_rate': float(htkout['sampling_rate']),
        'sample_size': htkout['sample_size'],
        'parameter_kind': htkout['parameter_kind'],
        'data': alldata
    }
